/*
  Anthyμ饤֥Υȥ롼
 */
#include <ctype.h>
#include <dic.h>
#include <conf.h>
#include <record.h>
#include "dic_ent.h"
#include "ddic.h"
#include "sdic.h"
#include "dic_cache.h"
#include "dic_session.h"

/* dummy node */
static struct seq_ent unkseq_ent_node;/*̤ʸ󤿤ȤбʸȤ*/
static struct seq_ent num_ent_node;/*ʤ*/
static struct seq_ent sep_ent_node;/*ѥ졼ʤɡ*/

static void init_dummy_nodes();
static sdic_t open_sdic();
static seq_ent_t get_extra_seq_ent_from_xstr(xstr *x);
static int get_extra_seq_ent_pos(struct seq_ent *,int );
static int get_extra_seq_ent_indep(struct seq_ent *);
static int get_extra_seq_ent_ct(struct seq_ent *,int ,int);
static int get_extra_seq_ent_wtype(struct seq_ent *,wtype_t );

void init_dummy_nodes()
{
  unkseq_ent_node.node_type = ST_UNKSEQ|ST_WORD;
  unkseq_ent_node.nr_dic_ents = 0;
  unkseq_ent_node.flags = F_NONE;
  num_ent_node.node_type = ST_WORD;
  num_ent_node.nr_dic_ents = 0;
  num_ent_node.flags = NF_NUM;
  sep_ent_node.node_type = ST_WORD|ST_SEP;
  sep_ent_node.nr_dic_ents = 0;
  sep_ent_node.flags = F_NONE;
}

sdic_t open_sdic()
{
  char *fn;
  fn = conf_get_str("DIC");
  if (!fn){
    return 0;
  }
  return create_sdic(fn);
}

seq_ent_t get_extra_seq_ent_from_xstr(xstr *x)
{
  if ( is_num_xstr(x)){
    return &num_ent_node;
  }
  /*ѿʤunkseq*/
  if ( is_ascii_xstr(x) ){
    return &unkseq_ent_node;
  }
  if ( x->len == 1 ){
    /*ˤΤäƤʤ1ʸʤ饻ѥ졼*/
    return &sep_ent_node;
  }
  return 0;
}

int init_anthy_dic()
{
  sdic_t si;
  conf_init();
  if ( init_sdic()){
    return -1;
  }
  init_wtypes();
  init_dummy_nodes();
  init_sessions();
  init_record();
  init_ext_ent();

  si = open_sdic();
  if ( !si ){
    return -1;
  }
  init_dic_cache(si);
  return 0;
}

seq_ent_t get_seq_ent_from_xstr(xstr *x)
{
  struct seq_ent *s;
  s = cache_get_seq_ent(x);
  if ( !s || s->nr_dic_ents==0 ){
    return get_extra_seq_ent_from_xstr(x);
  }
  return s;
}

int get_seq_ent_type(seq_ent_t s)
{
  struct seq_ent *se=s;
  if (!se){
    return ST_NONE;
  }
  return se->node_type;
}

int get_seq_flag(seq_ent_t se)
{
  if ( !se ){
    return 0;
  }
  return se->flags;
}

int get_nr_dic_ents(seq_ent_t se,xstr *xs)
{
  struct seq_ent *s = se;
  if ( !s ){
    return 0;
  }
  if ( s == &num_ent_node || 
       s == &unkseq_ent_node ||
       ( se->flags & NF_NUM)){
    return s->nr_dic_ents + get_nr_dic_ents_of_ext_ent(se,xs);
  }
  return s->nr_dic_ents;
}

int get_nth_dic_ent_str(seq_ent_t se,xstr *o,int n,xstr *x)
{
  struct seq_ent *s = se;
  if ( !s){
    return -1;
  }
  if ( n >= s->nr_dic_ents ){
    return get_nth_dic_ent_str_of_ext_ent(se,o,n - s->nr_dic_ents,x);
  }
  x->len = s->dic_ents[n]->str.len;
  x->str = xstr_dup_str(&s->dic_ents[n]->str);
  return 0;
}

int get_nth_dic_ent_freq(seq_ent_t se,int n)
{
  struct seq_ent *s = se;
  if (!s || s->nr_dic_ents <= n ){
      return 0;
  }
  return s->dic_ents[n]->freq;
}

int get_nth_dic_ent_wtype(seq_ent_t se,xstr *xs,int n,wtype_t *w)
{
  struct seq_ent *s = se;
  if (!s ){
    *w = wt_none;
    return -1;
  }
  if (s->nr_dic_ents <= n ){
    int r;
    r = get_nth_dic_ent_wtype_of_ext_ent(xs,n- s->nr_dic_ents,w);
    if ( r == -1 ){
      *w = wt_none;
    }
    return r;
  }
  *w =  s->dic_ents[n]->type;
  return 0;
}

int get_extra_seq_ent_pos(struct seq_ent *se,int pos)
{
  if ( se == &num_ent_node && pos == POS_NOUN){
    return 10;
  }
  if ( (se->node_type & ST_UNKSEQ) && pos == POS_NOUN){
    return 10;
  }
  return 0;
}

int get_seq_ent_pos(seq_ent_t se,int pos)
{
  int i,v=0;
  struct seq_ent *s = se;
  if (!s ){
    return 0;
  }
  if ( s->nr_dic_ents == 0 ){
    return get_extra_seq_ent_pos(se,pos);
  }
  for ( i = 0 ; i < s->nr_dic_ents ; i++ ){
    if ( wtype_get_pos(s->dic_ents[i]->type)== pos ){
      v += s->dic_ents[i]->freq;
      if ( v == 0){
	v = 1;
      }
    }
  }
  return v;
}

int get_extra_seq_ent_ct(struct seq_ent *se,int pos,int ct)
{
  if ( get_extra_seq_ent_pos(se,pos) && ct == CT_NONE){
    return 10;
  }
  return 0;
}

int get_seq_ent_ct(seq_ent_t se,int pos,int ct)
{
  int i,v=0;
  struct seq_ent *s = se;
  if (!s){
      return 0;
  }
  if ( s->nr_dic_ents == 0 ){
    return get_extra_seq_ent_ct(s,pos,ct);
  }
  for ( i = 0 ; i < s->nr_dic_ents ; i++ ){
    if ( wtype_get_pos(s->dic_ents[i]->type)== pos &&
	 wtype_get_ct(s->dic_ents[i]->type)==ct){
      v += s->dic_ents[i]->freq;
      if ( v == 0){
	v = 1;
      }
    }
  }
  return v;
}

int get_extra_seq_ent_wtype(struct seq_ent *se,wtype_t w)
{
  if ( wtype_get_pos(w) == POS_NOUN ){
    return 10;
  }
  return 0;
}

int get_seq_ent_wtype_freq(seq_ent_t se,wtype_t w)
{
  int i,f;
  struct seq_ent *s = se;
  if (!s){
      return 0;
  }
  if ( s->nr_dic_ents == 0 ){
    return get_extra_seq_ent_wtype(se,w);
  }
  f = 0;
  for ( i = 0 ; i < s->nr_dic_ents ; i++ ){
    if ( !wtypecmp(s->dic_ents[i]->type,w)){
      if ( f < s->dic_ents[i]->freq){
	f = s->dic_ents[i]->freq;
      }
    }
  }
  return f;
}

int get_extra_seq_ent_indep(struct seq_ent *se)
{
  if ( se == &num_ent_node || se == &unkseq_ent_node ){
    return 1;
  }
  return 0;
}

int get_seq_ent_indep(seq_ent_t se)
{
  int i;
  struct seq_ent *s = se;
  if (!s){
      return 0;
  }
  if ( s->nr_dic_ents == 0 ){
    return get_extra_seq_ent_indep(s);
  }
  for ( i = 0 ; i < s->nr_dic_ents ; i++ ){
    if ( wtype_get_indep(s->dic_ents[i]->type)){
      return 1;
    }
  }
  return 0;
}

dic_session_t dic_create_session()
{
  return create_session();
}

void dic_activate_session(dic_session_t d)
{
  activate_session(d);
}

void dic_release_session(dic_session_t d)
{
  release_session(d);
  shrink_cache();
}
