/*
  Comments in this program are written in Japanese.
  Because this program is Japanese input method.

  KanaKanji conversion engine Anthy.
  ̾Ѵ󥸥Anthy(󥷡)
 */
/*桼ץȤΥ󥿡ե
  Υδؿdll˥ݡȤƤ롣

  饤֥㴳̩˥顼å
 */
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include <dic.h>
#include <conf.h>
#include <anthy.h>
#include "main.h"
#include "context.h"

static int g_isInitOk;/*Anthyνλɤ*/

static int commit_all_segment_p(struct anthy_context *);

int anthy_init()
{
  if ( g_isInitOk ){
    return 0;
  }
  if ( init_anthy_dic() ){
    printf("failed to open dictionary.\n");
    return -1;
  }
  if ( init_seg_struct_tab() ){
    printf("failed to init segment table.\n");
    return -1;
  }
  if ( init_word_seq_tab() ){
    printf("failed to init word sequence table.\n");
    return -1;
  }
  g_isInitOk = 1;
  return 0;
}

void anthy_conf_override(char *var,char *val)
{
  conf_override(var,val);
}

int anthy_set_personality(char *id)
{
  return 0;
}

struct anthy_context *anthy_create_context()
{
  if (! g_isInitOk ){
    return 0;
  }
  return create_context();
}

void anthy_reset_context(struct anthy_context *ac)
{
  reset_context(ac);
}

void anthy_release_context(struct anthy_context *ac)
{
  release_context(ac);
}

void anthy_set_string(struct anthy_context *ac,char *s)
{
  xstr *x;
  x = cstr_to_xstr(s);
  context_set_str(ac,x);
  free_xstr(x);
}

void anthy_resize_segment(struct anthy_context *ac,int nth,int resize)
{
  context_resize_segment(ac,nth,resize);
}

int anthy_get_stat(struct anthy_context *ac,struct anthy_conv_stat *s)
{
  s->nr_segment = ac->seg_count;
  return 0;
}

int anthy_get_segment_stat(struct anthy_context *c,int n,struct anthy_segment_stat *s)
{
  struct seg_ent *se;
  se = get_nth_segment(c,n);
  if ( se){
    s->nr_candidate = se->cand_count;
    s->seg_len = se->str.len;
    return 0;
  }
  return -1;
}

int anthy_get_segment(struct anthy_context *ac,int nth_seg,int nth_cand,char *buf,int buflen)
{
  struct seg_ent *seg;
  char *p;
  if ( nth_seg < 0 || nth_seg >= ac->seg_count ){
    return -1;
  }
  seg = get_nth_segment(ac,nth_seg);
  if ( nth_cand < 0 || nth_cand >= seg->cand_count ){
    return -1;
  }
  p = xstr_to_cstr(&seg->cands[nth_cand]->str);
  if ( strlen(p)>buflen){
    free(p);
    return -1;
  }
  strcpy(buf,p);
  free(p);
  return 0;
}

int anthy_commit_segment(struct anthy_context *ac,int s,int c)
{
  struct seg_ent *seg;
  if ( s < 0 || s>= ac->seg_count ){
    return -1;
  }
  if ( commit_all_segment_p(ac)){
    /*ǤƤΥȤߥåȤƤ롣*/
    return -1;
  }
  seg = get_nth_segment(ac,s);
  seg->commit = c;
  if ( commit_all_segment_p(ac)){
    /*٤ƤΥȤߥåȤ줿*/
    proc_commit(ac);
  }
  return 0;
}

void anthy_print_context(struct anthy_context *ac)
{
  print_context(ac);
}

int commit_all_segment_p(struct anthy_context *ac)
{
  int i;
  struct seg_ent *se;
  for ( i = 0 ; i < ac->seg_count ; i++){
    se = get_nth_segment(ac,i);
    if ( se->commit <0 ){
      return 0;
    }
  }
  return 1;
}
