// Anthy libraryθƤӽФԤ
#include <stdlib.h>
#include "rkconv.h"
#include "keybind.h"
#include "anthyconv.h"

static bool isAnthyInitialized=false;

static void init_anthy_conv();

void init_anthy_conv()
{
  if ( isAnthyInitialized ){
    return ;
  }
  isAnthyInitialized = true;
}


char *jstring_to_str(jstring *s)
{
  jstring::iterator i;
  int l=0;
  char *c;
  for ( i = s->begin() ; i!= s->end() ; i++){
    l++;
    if ( *i > 256 ){
      l ++;
    }
  }
  c = (char *)malloc(l+1);
  c[l] = 0;
  l = 0;
  for ( i = s->begin() ; i!= s->end() ; i++){
    if ( *i < 256 ){
      c[l] = *i;
      l++;
    }else{
      c[l] = ((*i)>>8)|0x80;
      c[l+1] = ((*i)&0xff)|0x80;
      l+=2;
    }
  }
  return c;
}

void clear_jstring(jstring *js)
{
  js->erase(js->begin(),js->end());
}

void append_jstring(jstring *dst,jstring *src)
{
  jstring::iterator it;
  for ( it = src->begin() ;  it != src->end() ; it ++){
    dst->push_back(*it);
  }
}

void jstring_to_STRING(STRING *dst,jstring *src)
{
  char *s = jstring_to_str(src);
  *dst = createSTRING(s);
  free(s);
}

void str_to_jstring(jstring *js,char *str)
{
  int len,i;
  unsigned char *s = (unsigned char*)str;
  len = strlen(str);
  for ( i = 0 ; i < len ; i ++ ){
    jchar ch;
    if ( s[i] &0x80 ){
      ch = (s[i]<<8)|(s[i+1]);
      i++;
    }else{
      ch = s[i];
    }
    js->push_back(ch);
  }
}

void print_jchar(jchar c)
{
  char buf[3];
  if ( c < 256 ){
    buf[0] = c;
    buf[1] = 0;
  }else{
    buf[2] = 0;
    buf[1] = (c & 255)| 0x80;
    buf[0] = ((c>>8) & 255)| 0x80;
  }
  printf(buf);
}

void print_jstring(jstring *js)
{
  jstring::iterator it;
  for ( it = js->begin() ; it != js->end() ; it++){
    print_jchar(*it);
  }
  printf("\n");
}

AnthyConv::AnthyConv(anthy_context_t ac)
{
  mAC = ac;
  mRkConv = new RKConv();
  mMode = INPUT_MODE;
  mRkConv->select_map(rk_hira_map);
}

AnthyConv::~AnthyConv()
{
}

int AnthyConv::proc_key(keyEvent *ev)
{
  switch(mMode ){
  case RAW_MODE:
    return proc_raw_mode(ev);
  case INPUT_MODE:
    return proc_input_mode(ev);
  case COMPOSE_MODE:
    return proc_compose_mode(ev);
  }
  return 0;
}

int AnthyConv::proc_raw_mode(keyEvent *ev)
{
  if ( ev->keychar == ' ' ){
  }
  return COMMIT_RAW;
}

int AnthyConv::proc_input_mode(keyEvent *ev)
{
  if ( isBindTo(ev,A_do_conv) ){
    mRkConv->flush();
    do_conv();
    return UPDATE_PREEDIT;
  }
  if (isBindTo(ev,A_delete_back) ){
    if ( !mRkConv->back_space() ){
      if ( mStr.size() ){
	mStr.pop_back();
      }else{
	return COMMIT_RAW;
      }
    }
    return UPDATE_PREEDIT;
  }
  if ( isBindTo(ev,A_cancel )){
    mRkConv->flush();
    clear_jstring(&mStr);
    return UPDATE_PREEDIT;
  }
  mRkConv->push_key(ev->keychar);

  jchar ch;
  ch = mRkConv->get_char();
  while(ch ){
    mStr.push_back(ch);
    ch = mRkConv->get_char();
  }
  return UPDATE_PREEDIT;
}

int AnthyConv::proc_compose_mode(keyEvent *ev)
{
  if ( isBindTo(ev,A_next_segment) && mCurSeg+1 < mSegs.size()){
    mCurSeg++;
    return UPDATE_PREEDIT;
  }
  if ( isBindTo(ev,A_prev_segment) && mCurSeg > 0){
    mCurSeg--;
    return UPDATE_PREEDIT;
  }
  if ( isBindTo(ev,A_expand_segment)){
    anthy_resize_segment(mAC,mCurSeg,1);
    update_anthy_stat(mCurSeg);
    return UPDATE_PREEDIT;
  }
  if ( isBindTo(ev,A_shorten_segment)){
    anthy_resize_segment(mAC,mCurSeg,-1);
    update_anthy_stat(mCurSeg);
    return UPDATE_PREEDIT;
  }
  if ( isBindTo(ev,A_next_candidate)){
    if ( mSegs[mCurSeg].cur_cand < mSegs[mCurSeg].cands.size()){
      mSegs[mCurSeg].cur_cand ++;
    }
    if ( mSegs[mCurSeg].cur_cand == mSegs[mCurSeg].cands.size()-1){
      mSegs[mCurSeg].cur_cand = 0;
    }
    return UPDATE_PREEDIT;
  }
  if ( isBindTo(ev,A_cancel) ){
    mMode = INPUT_MODE;
    return UPDATE_PREEDIT;
  }
  if ( isBindTo(ev,A_commit) ){
    clear_jstring(&mCommit);
    clear_jstring(&mStr);
    vector<segment>::iterator it;
    for ( it = mSegs.begin() ; it != mSegs.end() ; it ++){
      append_jstring(&mCommit,&(*it).cands[(*it).cur_cand]);
    }
    mMode = INPUT_MODE;
    do_commit();
    return COMMIT_PENDING|UPDATE_PREEDIT;
  }
  return UPDATE_PREEDIT;
}

STRING AnthyConv::get_pending()
{
  STRING str;
  jstring_to_STRING(&str,&mCommit);
  return str;
}

STRING AnthyConv::get_status()
{
  return createSTRING("Anthy");
}

void AnthyConv::do_conv()
{
  char *str;
  str = jstring_to_str(&mStr);
  anthy_set_string(mAC,str);
  free(str);
  mMode = COMPOSE_MODE;
  update_anthy_stat(0);
  mCurSeg = 0;
}

void AnthyConv::do_commit()
{
  int i;
  for ( i = 0 ; i < mSegs.size() ; i ++ ){
    anthy_commit_segment(mAC,i,mSegs[i].cur_cand);
  }
}

void AnthyConv::get_preedit(preedit *p)
{
  p->segs.erase(p->segs.begin(),p->segs.end());
  if ( mMode == COMPOSE_MODE ){
    int i;
    for ( i = 0 ; i < mSegs.size() ; i++){
      pe_string ps;
      jstring_to_STRING(&ps.str,&mSegs[i].cands[mSegs[i].cur_cand]);
      ps.stat = NORMAL_VIDEO;
      if ( mCurSeg == i ){
	ps.stat = REVERSE_VIDEO;
      }
      p->segs.push_back(ps);
    }
  }else if ( mMode == INPUT_MODE ){
    jstring js=mStr;
    char buf[10];
    mRkConv->get_pending_char(buf,10);
    for ( int i = 0 ; i < strlen(buf) ; i++){
      js.push_back(buf[i]);
    }
    print_jstring(&js);
    pe_string ps;
    jstring_to_STRING(&ps.str,&js);
    p->segs.push_back(ps);
  }
}

void AnthyConv::update_anthy_stat(int s)
{
  while(mSegs.size()> s ){
    mSegs.pop_back();
  }
  struct anthy_conv_stat cs;
  int i,j;
  anthy_get_stat(mAC,&cs);
  for  ( i = s ; i < cs.nr_segment ; i++){
    segment seg;
    anthy_segment_stat ss;
    anthy_get_segment_stat(mAC,i,&ss);
    seg.cur_cand = 0;
    for ( j = 0 ; j < ss.nr_candidate ; j++ ){
      char buf[256];
      anthy_get_segment(mAC,i,j,buf,256);
      jstring js;
      str_to_jstring(&js,buf);
      //print_jstring(&js);
      seg.cands.push_back(js);
    }
    mSegs.push_back(seg);
  }
}

AnthyConv *create_anthy_context()
{
  anthy_context_t ac;
  if ( isAnthyInitialized ){
    init_anthy_conv();
    return 0;
  }
  ac = anthy_create_context();
  if ( !ac ){
    return 0;
  }
  return new AnthyConv(ac);
}
